/*
Module : DTime+CDate.H
Purpose: Defines the interface for the DTime+ CDate class
Created: PJN / 01-05-2006
History: PJN / 02-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.

Copyright (c) 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#ifndef __DTIMEPLUSCDATE_H__
#define __DTIMEPLUSCDATE_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////// Includes //////////////////////////////////////////

#include "DTime+Defines.h"


/////////////////////////// Classes ///////////////////////////////////////////

namespace DTimePlus
{

//A Date class with a granularity of 1 Day in the Gregorian Calendar
template <class TCALENDAR = CGregorianCalendar>
class CDate : public TCALENDAR
{
public:
//Constructors
  CDate()
  {
    Set();
  }

  CDate(long Year, WORD month, WORD Day, bool bDoAssert = true)
  { 
    Set(Year, month, Day, bDoAssert);
  }

  CDate(const SYSTEMTIME& SystemTime, bool bDoAssert = true)
  {
    Set(SystemTime, bDoAssert);
  }

  CDate(const FILETIME& FileTime, bool bDoAssert = true)
  {
    Set(FileTime, bDoAssert);
  }

  CDate(long Year, WORD month, WORD WeekOfMonth, DayOfWeek dow, bool bDoAssert = true)
  {
    Set(Year, month, WeekOfMonth, dow, bDoAssert);
  }

  CDate(long Days, DateEpoch e, bool bDoAssert = true)
  {
    Set(Days, e, bDoAssert);
  }

  CDate(const CDate& d)
  {
    m_lDays = d.m_lDays;
  }

  CDate(const CTime& ctime, bool bDoAssert = true)
  {
    Set(ctime, bDoAssert);
  }
                   
  #ifndef DTIME_NO_OLE_SUPPORT                  
  CDate(const COleDateTime& oleTime)
  {
    Set(oleTime);
  }                     
  #endif

  CDate(const CString& sDate, const CString& sFormat, bool bDoAssert = true)
  {
    Set(sDate, sFormat, bDoAssert);
  }

  //Set operators
  CDate& Set()
  {
    //Create an invalid value
    m_lDays = LONG_MIN;

    return *this;
  }

  CDate& Set(long Year, WORD month, WORD Day, bool bDoAssert = true)
  {
    if (!TCALENDAR::DMYToLong(Year, month, Day, m_lDays, bDoAssert))
      Set();
 
    return *this;
  }

#ifdef _DEBUG
  CDate& Set(long Year, WORD month, WORD WeekOfMonth, DayOfWeek dow, bool bDoAssert = true)
#else
  CDate& Set(long Year, WORD month, WORD WeekOfMonth, DayOfWeek dow, bool /*bDoAssert*/ = true)
#endif
  {
    //Validate our parameters
    if (WeekOfMonth < 1 || WeekOfMonth > 5 || dow < Monday && dow > Sunday)
    {
    #ifdef _DEBUG
      if (bDoAssert)
        ASSERT(false);
    #endif
      Set();
      return *this;
    }

    if (WeekOfMonth < 5)
    {
      CDate FirstMonth(Year, month, 1);
      if (!FirstMonth.IsValid())
      {
      #ifdef _DEBUG
        if (bDoAssert)
          ASSERT(false);
      #endif
        Set();
        return *this;
      }
      WORD dow1 = static_cast<WORD>(FirstMonth.GetDayOfWeek());
      WORD Day = static_cast<WORD>(((dow - dow1 + 7) % 7) + (7*(WeekOfMonth-1)) + 1);
      Set(Year, month, Day);
    }
    else  //5 means the last week of the month
    {
      WORD NewMonth = static_cast<WORD>(month + 1);
      long NewYear = Year;
      if (NewMonth > 12)
      {
        NewMonth = 1;
        ++NewYear;
      }
       
      Set(NewYear, NewMonth, 1, dow);
      AddWeek(-1);
    }

    return *this;
  }

#ifdef _DEBUG
  CDate& Set(long Days, DateEpoch e, bool bDoAssert = true)
#else
  CDate& Set(long Days, DateEpoch e, bool /*bDoAssert*/ = true)
#endif
  {
    switch (e)
    {
      case EPOCH_GREG:
      {
        m_lDays = Days;
        break;
      }
      case EPOCH_JD:    
      {
        m_lDays = Days - 2299161; //-2299161 coresponds to the GDN of JD Epoch namely 1 January -4712 
        break;
      }
      case EPOCH_MJD:   
      {
        m_lDays = Days + 100839; //100839 coresponds to the GDN of MJD Epoch namely 17 November 1858 
        break;
      }
      case EPOCH_1900:
      {
        m_lDays = Days + 115860; //115860 corresponds to the GDN of 1 January 1900
        break;
      }
      case EPOCH_1950:
      {
        m_lDays = Days + 134122; //134122 corresponds to the GDN of 1 January 1950
        break;
      }
      case EPOCH_CTIME:
      {
        m_lDays = Days + 141427; //141427 corresponds to the GDN of 1 January 1970
        break;
      }
      case EPOCH_2000:  
      {
        m_lDays = Days + 152384; //152384 corresponds to the GDN of 1 January 2000
        break;
      }
      default: 
      {
        ASSERT(false); //should not occur
        break;
      }
    }

    if (m_lDays > 2146905911)  //Largest valid GDN
    {
    #ifdef _DEBUG  
      if (bDoAssert)
        ASSERT(false);
    #endif
      Set();
    }

    return *this;
  }

  CDate& Set(const SYSTEMTIME& SystemTime, bool bDoAssert = true)
  {
    Set(SystemTime.wYear, SystemTime.wMonth, SystemTime.wDay, bDoAssert);
    return *this;
  }

  CDate& Set(const FILETIME& FileTime, bool bDoAssert = true)
  {
    SYSTEMTIME st;
    if (::FileTimeToSystemTime(&FileTime, &st))
      Set(st);
    else
    {
      if (bDoAssert)
        ASSERT(FALSE);
      Set();
    }

    return *this;
  }

  CDate& Set(const CTime& ctime, bool bDoAssert = true)
  {
    tm TM;
    if (ctime.GetGmtTm(&TM))
      Set(TM.tm_year + 1900, static_cast<WORD>(TM.tm_mon + 1), static_cast<WORD>(TM.tm_mday), bDoAssert);
    else
    {
      if (bDoAssert)
        ASSERT(FALSE);

      Set();
    }

    return *this;
  }
      
  #ifndef DTIME_NO_OLE_SUPPORT
  CDate& Set(const COleDateTime& oleTime)
  {
    CDate rVal;
    if (oleTime.GetStatus() == COleDateTime::valid)
      Set(static_cast<long>(oleTime.GetYear()), static_cast<WORD>(oleTime.GetMonth()), static_cast<WORD>(oleTime.GetDay()));
    else
      Set();
    return *this;
  }                     
  #endif

  CDate& SetFromCollate(long Date, bool bDoAssert = true)
  {
    CString sDate;
    sDate.Format(_T("%ld"), Date);

    int nLen = sDate.GetLength();
    CString sDay = sDate.Mid(nLen-2, 2);
    CString sMonth = sDate.Mid(nLen-4, 2);
    CString sYear = sDate.Left(nLen-4);

    WORD Day = static_cast<WORD>(_ttoi(sDay));
    WORD month = static_cast<WORD>(_ttoi(sMonth));
    long Year = _ttol(sYear); 
  
    return Set(Year, month, Day, bDoAssert);
  }

  /*
  sDate is a legal date string as defined by sFormat.

  sFormat is a 5 character string where the first 3 indicate the
  order of the month/day/year, the fourth character indicates the length
  of the year, and the fifth character indicates the separator.

  For example:

  sFormat	  sDate          Result

  "dmy2/"   "20/03/96"     20th March 1996
  "mdy4-"   "03-28-1996"   28th March 1996

  This feature allows for the creation of CDate's from a string
  as entered in a masked edit field or from a parsed report.
  */
  CDate& Set(const CString& sDate, const CString& sFormat, bool bDoAssert = true)
  {
    if (sFormat.GetLength() != 5)
    {
      if (bDoAssert)
        ASSERT(false); //sFormat string provided was invalid!
      Set();
    }
    else
    {
      WORD Day = 0;
      WORD month = 1;
      long Year = 0;

      CString sTokens(sFormat.GetAt(4));
      CString sD(sDate);
      LPTSTR pszDate = sD.GetBuffer(sD.GetLength());
    #if (_MSC_VER >= 1400)
      TCHAR* pszNextToken;
      LPTSTR pszToken = _tcstok_s(pszDate, sTokens, &pszNextToken);
    #else  
      LPTSTR pszToken = _tcstok(pszDate, sTokens);
    #endif

      for (int i=0; i<3; i++)
      {
        ASSERT(pszToken);
        switch (sFormat[i])
        {
          case _T('d'): 
          {
            Day = static_cast<WORD>(_ttoi(pszToken)); 
            break;
          }
          case _T('m'): 
          {
            month = static_cast<WORD>(_ttoi(pszToken)); 
            break;
          }
          case _T('y'): 
          {
            Year  = _ttol(pszToken); 
            TCHAR c = sFormat.GetAt(3);
            ASSERT((c == _T('2')) || (c == _T('4')));
            if (c == _T('2'))
            {
              if (Year >= 50)
                Year += 1900;
              else if (Year < 50)
                Year += 2000;
            }
            break;
          }
          default: 
          {
            ASSERT(false); 
            break;
          }
        }

      #if (_MSC_VER >= 1400)
        pszToken = _tcstok_s(NULL, sTokens, &pszNextToken);
      #else
        pszToken = _tcstok(NULL, sTokens);
      #endif
      }

      sD.ReleaseBuffer();

      Set(Year, month, Day, bDoAssert);
    }

    return *this;
  }

//Static constructors
  static CDate CurrentLocalDate(bool bDoAssert = true)
  {   
    //PC clock will always return time in Gregorian calendar
    SYSTEMTIME lt;
    GetLocalTime(&lt);
    CDate<CGregorianCalendar> date(lt, bDoAssert);
  
    if (date.IsValid())
      return CDate(date.GetValue(), EPOCH_GREG);
    else
      return CDate();
  }

  static CDate CurrentSystemDate(bool bDoAssert = true)
  {   
    //PC clock will always return time in Gregorian calendar
    SYSTEMTIME st;
    ::GetSystemTime(&st);
    CDate<CGregorianCalendar> date(st, bDoAssert);
  
    if (date.IsValid())
      return CDate(date.GetValue(), EPOCH_GREG);
    else
      return CDate();
  }

  static CDate JDEpoch()
  {
    return CDate(0, EPOCH_JD);
  }

  static CDate MJDEpoch()
  {
    return CDate(0, EPOCH_MJD);
  }

  static CDate Epoch1900()
  {
    return CDate(0, EPOCH_1900);
  }

  static CDate Epoch1950()
  {
    return CDate(0, EPOCH_1950);
  }

  static CDate EpochCTime()
  {
    return CDate(0, EPOCH_CTIME);
  }

  static CDate Epoch2000()
  {
    return CDate(0, EPOCH_2000);
  }

  static CDate GregorianEpoch()
  {
    return CDate(0, EPOCH_GREG);
  }                 

//Static operations
  static CString GetWindowsFormatString(LCID Locale = LOCALE_USER_DEFAULT)
  {
    //Is there to be a leading 0 on the day?
    bool bDayLZ = false;
    CString sBuf = SafeGetLocaleInfo(Locale, LOCALE_IDAYLZERO);
    if (sBuf == _T("0"))  //No leading zeros for days
      bDayLZ = false;
    else if (sBuf == _T("1"))  //Leading zeros for days
      bDayLZ = true;
    else
      ASSERT(false);

    //Is there to be a leading 0 on the month?
    bool bMoLZ = true;
    sBuf = SafeGetLocaleInfo(Locale, LOCALE_IMONLZERO);
    if (sBuf == _T("0"))  //No leading zeros for months
      bMoLZ = false;
    else if (sBuf == _T("1"))  //Leading zeros for months
      bMoLZ = true;
    else
      ASSERT(false);

    //Find out if only 2 digits are to be displayed for the year
    bool b2DigitYear = false;
    sBuf = SafeGetLocaleInfo(Locale, LOCALE_ICENTURY);
    if (sBuf == _T("0"))  //Abbreviated 2-digit century
      b2DigitYear = true;
    else if (sBuf == _T("1"))  //Full century
      b2DigitYear = false;
    else
      ASSERT(false);
  
    //Get the date separator character
    CString sSep(_T("/"));
    sBuf = SafeGetLocaleInfo(Locale, LOCALE_SDATE);
    if (sBuf.GetLength())
      sSep = sBuf;

    //Retrieve the short date formatting string
    sBuf = SafeGetLocaleInfo(Locale, LOCALE_IDATE);

    //What will be the return value from this function
    CString sFormatString;

    //Format according to the short date formatting string
    if (sBuf == _T("0"))  //MDY format
    {
      if (b2DigitYear)
        sFormatString.Format(_T("%s%s%s%s%%y"), bMoLZ ? _T("%m") : _T("%#m"), sSep, bDayLZ ? _T("%d") : _T("%#d"), sSep);
      else
        sFormatString.Format(_T("%s%s%s%s%%Y"), bMoLZ ? _T("%m") : _T("%#m"), sSep, bDayLZ ? _T("%d") : _T("%#d"), sSep);
    }
    else if (sBuf == _T("1"))  //DMY format
    {
      if (b2DigitYear)
        sFormatString.Format(_T("%s%s%s%s%%y"), bDayLZ ? _T("%d") : _T("%#d"), sSep, bMoLZ ? _T("%m") : _T("%#m"), sSep);
      else
        sFormatString.Format(_T("%s%s%s%s%%Y"), bDayLZ ? _T("%d") : _T("%#d"), sSep, bMoLZ ? _T("%m") : _T("%#m"), sSep);
    }
    else 
    {
    #ifdef _DEBUG
      if (sBuf != _T("2"))  //No Recognized format available
        ASSERT(false);
    #endif

      //YMD Format
      if (b2DigitYear)
        sFormatString.Format(_T("%%y%s%s%s%s"), sSep, bMoLZ ? _T("%m") : _T("%#m"), sSep, bDayLZ ? _T("%d") : _T("%#d"));
      else
        sFormatString.Format(_T("%%Y%s%s%s%s"), sSep, bMoLZ ? _T("%m") : _T("%#m"), sSep, bDayLZ ? _T("%d") : _T("%#d"));
    }

    return sFormatString;
  }

//Operations
  WORD DaysSinceFirstOfYear() const
  {
    ASSERT(IsValid());

    return static_cast<WORD>(*this - CDate(GetYear(), 1, 1));
  }

  WORD DaysSinceEndPreviousYear() const
  {
    ASSERT(IsValid());

    return static_cast<WORD>(DaysSinceFirstOfYear() + 1);
  }

  WORD GetDay() const
  {
    ASSERT(IsValid());

    return GetDate().wDay;
  }

  WORD GetMonth() const
  {
    ASSERT(IsValid());

    return GetDate().wMonth;
  }

  long GetYear() const
  {
    ASSERT(IsValid());

    return GetDate().lYear;
  }

  long GetCEBCEYear(bool& IsCE) const
  {
    ASSERT(IsValid());

    long Year = GetDate().lYear;

    if (Year > 0)
    {
      IsCE = true;
      return Year;
    }
    else
    {
      IsCE = false;
      return labs(GetYear() - 1);
    }
  }

  static WORD Get2DigitYear(long Year)
  {
    Year = labs(Year);
    return static_cast<WORD>(Year - ((Year/100)*100));
  }

  WORD Get2DigitYear() const
  {
    ASSERT(IsValid());

    return Get2DigitYear(GetYear());
  }

  long JD() const
  {
    ASSERT(IsValid());

    return *this - JDEpoch();
  }

  long GetValue() const
  {
    ASSERT(IsValid());

    return m_lDays;
  }

  DateS GetDate() const
  {
    ASSERT(IsValid());

    return TCALENDAR::GetDate(m_lDays);
  }

  bool IsLeap() const
  {
    ASSERT(IsValid());

    return TCALENDAR::IsLeap(GetYear());
  }

  WORD DaysInYear() const
  {
    ASSERT(IsValid());
 
    return TCALENDAR::DaysInYear(GetYear());
  }

  WORD DaysInMonth() const
  {
    ASSERT(IsValid());

    DateS ds(GetDate());

    return TCALENDAR::DaysInMonth(ds.lYear, ds.wMonth);
  }

  void AddYear(int Years)
  {
    ASSERT(IsValid());

    long Months = Years*12;
    if (Months >= INT_MIN && Months <= INT_MAX)
      AddMonth(static_cast<int>(Months));
    else
    {
      ASSERT(false);
      Set();
    }
  }

  void AddMonth(int Months)
  {
    ASSERT(IsValid());

    int NewMonth = GetMonth();
    long NewYear = GetYear();
    NewMonth += Months;
    if ((NewMonth > 12) || (NewMonth < 1))
    {
      NewYear += NewMonth / 12;
      NewMonth = NewMonth % 12;
    }
    if (NewMonth < 1)
    {
      --NewYear;
      NewMonth += 12;
    }

    WORD Day = GetDay();

    WORD MaxDays = TCALENDAR::DaysInMonth(NewYear, static_cast<WORD>(NewMonth));
    if (Day > MaxDays)
      Day = MaxDays;
    Set(NewYear, static_cast<WORD>(NewMonth), Day);
  }

  void AddWeek(int Weeks)
  {
    ASSERT(IsValid());

    *this += 7L*(static_cast<long>(Weeks));
  }

  DayOfWeek GetDayOfWeek() const
  {
    ASSERT(IsValid());

    return static_cast<DayOfWeek>(lmod(m_lDays+12, 7) + 1);
  }

  CString GetFullStringDayOfWeek(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());

    return TCALENDAR::GetFullStringDayOfWeek(GetDayOfWeek(), Locale);
  }

  CString GetAbrStringDayOfWeek(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());

    return TCALENDAR::GetAbrStringDayOfWeek(GetDayOfWeek(), Locale);
  }

  CString GetFullStringMonth(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());

    return TCALENDAR::GetFullStringMonth(GetMonth(), Locale);
  }

  CString GetAbrStringMonth(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());

    return TCALENDAR::GetAbrStringMonth(GetMonth(), Locale);
  }

  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const
  {
    if (!IsValid())
      return false;

    DateS ds(GetDate());

    //If our date cannot fit into the range allowed by SYSTEMTIME
    if ((ds.lYear < 0) || (ds.lYear > USHRT_MAX))
      return false;

    SystemTime.wYear = static_cast<WORD>(ds.lYear);
    SystemTime.wMonth = ds.wMonth;
    SystemTime.wDayOfWeek = static_cast<WORD>(ds.weekday - 1);  //SYSTEMTIME uses 0 based indices
    SystemTime.wDay = ds.wDay;
    SystemTime.wHour = 0;
    SystemTime.wMinute = 0;
    SystemTime.wSecond = 0;
    SystemTime.wMilliseconds = 0;

    return true;
  }

  bool GetAsFileTime(FILETIME& FileTime) const
  {
    SYSTEMTIME SystemTime;
    BOOL bSuccess = GetAsSystemTime(SystemTime);
    bSuccess = bSuccess && ::SystemTimeToFileTime(&SystemTime, &FileTime);
    return (bSuccess != 0);
  }
                          
  bool GetAsTM(tm& time) const
  {
    //What will be the return value from this function (assume the best)
    bool rVal = true;

    if (!IsValid())
      return false;
  
    long MinYear = 1900L + INT_MIN;
    long MaxYear = INT_MAX;

    //handle range errors
    DateS ds(GetDate()); 
    if ((ds.lYear < MinYear) || (ds.lYear > MaxYear))
      rVal = false;
    else
    {
      time.tm_sec   = 0;
      time.tm_min   = 0;
      time.tm_hour  = 0;
      time.tm_isdst = 0;
      time.tm_year  = static_cast<int>(ds.lYear - 1900);
      time.tm_mon   = ds.wMonth - 1; //tm struct uses 0 based indices
      time.tm_wday  = ds.weekday - 1; //tm struct uses 0 based indices
      time.tm_mday  = ds.wDay; 
      time.tm_yday  = ds.wYearday - 1; //Returns days since Jan 1
    }
              
    return rVal;
  }

  WORD GetWeekOfYear(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());
               
    CDate FirstOfYear(GetYear(), 1, 1, GetBeginingDayOfWeek(Locale));
    long diff = *this - FirstOfYear;
    if (diff < 0)
      return 0;
    else
      return static_cast<WORD>(diff/7 + 1);
  }

  WORD GetWeekOfMonth(LCID Locale = LOCALE_USER_DEFAULT) const
  {
    ASSERT(IsValid());
                  
    CDate FirstOfMonth(GetYear(), GetMonth(), 1, GetBeginingDayOfWeek(Locale));
    long diff = *this - FirstOfMonth;
    if (diff < 0)
      return 0;
    else
      return static_cast<WORD>(diff/7 + 1);
  }

  bool IsValid() const
  {
    return (m_lDays != LONG_MIN);
  }

  long Collate() const
  {
    DateS s(GetDate());
    return (s.lYear*10000) + (s.wMonth*100) + s.wDay;
  }

  CDate<CJulianCalendar> AsJulianCDate() const
  {
    CDate<CJulianCalendar> rVal;
    rVal.Set(m_lDays, EPOCH_GREG);

    return rVal;
  }

  CDate<CGregorianCalendar> AsGregorianCDate() const
  {
    CDate<CGregorianCalendar> rVal;
    rVal.Set(m_lDays, EPOCH_GREG);

    return rVal;
  }

  CDate<CHebrewCalendar> AsHebrewCDate() const
  {
    CDate<CHebrewCalendar> rVal;
    rVal.Set(m_lDays, EPOCH_GREG);

    return rVal;
  }
                                    
//Creation of other CDate's from this instance
  CDate FirstThisMonth() const
  {
    DateS s(GetDate());
    return CDate(s.lYear, s.wMonth, 1);  
  }

  CDate LastThisMonth() const
  {
    DateS s(GetDate());
    return CDate(s.lYear, s.wMonth, TCALENDAR::DaysInMonth(s.lYear, s.wMonth));  
  }

  CDate FirstThisYear() const
  {
    DateS s(GetDate());
    CDate d(s.lYear, 1, 1);
    return d;  
  }

  CDate LastThisYear() const
  {
    DateS s(GetDate());
    CDate d(s.lYear, 12, TCALENDAR::DaysInMonth(s.lYear, 12));
    return d;  
  }

//Overloaded Arithmetic Operators
  CDate& operator=(const CDate& d)
  {
    m_lDays = d.m_lDays;
    return *this;
  }

  CDate operator+(long Days) const
  {
    if (IsValid())
      return CDate(GetValue() + Days, EPOCH_GREG);
    else
      return CDate();
  }

  long operator-(const CDate& d) const
  {
    if (IsValid() & d.IsValid())
      return m_lDays - d.m_lDays;
    else
      return 0L;
  }

  CDate operator-(long Days) const
  {
    if (IsValid())
      return CDate(GetValue() - Days, EPOCH_GREG);
    else
      return CDate();
  }

  CDate& operator+=(long Days)
  {
    *this = *this + Days;
    return *this;
  }

  CDate& operator-=(long Days)
  {
    *this = *this - Days;
    return *this;
  }

  CDate& operator++()
  {
    *this = *this + 1;
    return *this;
  }

  CDate& operator--()
  {
    *this = *this - 1;
    return *this;
  }

//Overloaded Equality operators
  bool operator==(const CDate& d) const
  {
    bool bThisValid = IsValid();
    bool bOtherValid = d.IsValid();

    if (bThisValid && bOtherValid)
      return (m_lDays == d.m_lDays);
    else
      return (bThisValid == bOtherValid);
  }

  bool operator>(const CDate& d) const
  {
    bool bThisValid = IsValid();
    bool bOtherValid = d.IsValid();
    if (bThisValid && bOtherValid)
      return m_lDays > d.m_lDays;
    else
      return false; //Only makes sense to compare 2 valid valuess
  }

  bool operator>=(const CDate& d) const
  {
    return operator==(d) || operator>(d);
  }

  bool operator<(const CDate& d) const
  {
    bool bThisValid = IsValid();
    bool bOtherValid = d.IsValid();
    if (bThisValid && bOtherValid)
      return m_lDays < d.m_lDays;
    else
      return false; //Only makes sense to compare 2 valid valuess
  }

  bool operator<=(const CDate& d) const
  {
    return operator==(d) || operator<(d);
  }

  bool operator!=(const CDate& d) const
  {
    return !operator==(d);
  }

//Display
  /*    //The Following Format parameters are supported
  %a    Abbreviated weekday name
  %A    Full weekday name
  %b    Abbreviated month name
  %B    Full month name
  %d    Day of month as decimal number (01 - 31)
  %j    Day of year as decimal number (001 - 366)
  %m    Month as decimal number (01 - 12)
  %U    Week of year as decimal number
  %w    Weekday as decimal number (1 - 7; Sunday is 1)
  %x    Windows Short date representation using GetDateFormat() Win32 API
  %y    Year without century, as decimal number (00 - 99)
  %Y    Year with century, as decimal number
  %#x   Windows Long date representation using GetDateFormat() Win32 API
  %#d, %#j, %#m, %#U, %#y Remove leading zeros (if any)
  */
  CString Format(const CString& sFormat = GetWindowsFormatString(LOCALE_USER_DEFAULT), LCID Locale = LOCALE_USER_DEFAULT) const
  { 
    //Cache the date values for performance reasons
    DateS ds = GetDate();

    //What will be the return value from this function
    CString rVal;

    if (IsValid())
    {
      CString sBuffer;
      int sFmtLength = sFormat.GetLength();

      for (int i=0; i<sFmtLength; i++)
      {
        TCHAR c = sFormat.GetAt(i);
        if (c == _T('%'))
        {
          ++i;
          if (i < sFmtLength)
          {
            c = sFormat.GetAt(i);
            switch (c)
            {
              case _T('a'):
              {
                rVal += GetAbrStringDayOfWeek(Locale);
                break;
              }
              case _T('A'):
              {
                rVal += GetFullStringDayOfWeek(Locale);
                break;
              }
              case _T('b'):
              {
                rVal += GetAbrStringMonth(Locale);
                break;
              }
              case _T('B'):
              {
                rVal += GetFullStringMonth(Locale);
                break;
              }
              case _T('d'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(ds.wDay));
                rVal += sBuffer;
                break;
              }
              case _T('j'):
              {
                sBuffer.Format(_T("%.03d"), static_cast<int>(ds.wYearday));
                rVal += sBuffer;
                break;
              }
              case _T('m'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(ds.wMonth));
                rVal += sBuffer;
                break;
              }
              case _T('U'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(GetWeekOfYear(Locale)));
                rVal += sBuffer;
                break;
              }
              case _T('w'):
              {
                sBuffer.Format(_T("%d"), static_cast<int>(ds.weekday));
                rVal += sBuffer;
                break;
              }                                   
              case _T('x'):
              {
                sBuffer.Empty();
                SYSTEMTIME SystemTime;
                if (GetAsSystemTime(SystemTime))
                  rVal += SafeGetDateFormat(Locale, DATE_SHORTDATE, &SystemTime);
                else
                  TRACE(_T("Failed in call to SafeGetDateFormat in CDate::Format()\n"));
                break;
              }                 
              case _T('y'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(Get2DigitYear(ds.lYear)));
                rVal += sBuffer;
                break;
              }
              case _T('Y'):
              {
                sBuffer.Format(_T("%ld"), ds.lYear);
                rVal += sBuffer;
                break;
              }
              case _T('#'):
              {
                if (i < sFmtLength)
                {
                  ++i;
                  c = sFormat.GetAt(i);
                  switch (c)
                  {
                    case _T('d'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wDay));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('j'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wYearday));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('m'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wMonth));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('U'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(GetWeekOfYear(Locale)));
                      rVal += sBuffer;
                      break;
                    }                                    
                    case _T('x'):
                    {
                      sBuffer.Empty();
                      SYSTEMTIME SystemTime;
                      if (GetAsSystemTime(SystemTime))
                        rVal += SafeGetDateFormat(Locale, DATE_LONGDATE, &SystemTime);
                      else
                        TRACE(_T("Failed in call to SafeGetDateFormat in CDate::Format()\n"));
                      break;
                    }               
                    case _T('y'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(Get2DigitYear(ds.lYear)));
                      rVal += sBuffer;
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }

    return rVal;
  }

//Serialization
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      WORD wVersion = 0x101;
      ar << wVersion;

      ar << m_lDays;
    }
    else
    {
      WORD wVersion;
      ar >> wVersion;
      ASSERT(wVersion == 0x100 || wVersion == 0x101);

      ar >> m_lDays;
      if (wVersion == 0x100)
      {
        bool bOldInGregCalendar;
        ar.Read(&bOldInGregCalendar, sizeof(bOldInGregCalendar));
      }
    }
  }

  friend CArchive& operator<<(CArchive& ar, CDate& Date)
  {
    ASSERT(ar.IsStoring());
    Date.Serialize(ar);
    return ar;
  }

  friend CArchive& operator>>(CArchive& ar, CDate& Date)
  {
    ASSERT(!ar.IsStoring());
    Date.Serialize(ar);
    return ar;
  }

protected:
//Member variables
  long m_lDays; //Count of days since 15 October 1582 (Gregorian Calendrical System or n.s) 
};

} //namespace DTimePlus

#endif //__DTIMEPLUSCDATE_H__
