/*
Module : DTime+CLDate.H
Purpose: Defines the interface for the DTime+ CLDate class
Created: PJN / 01-05-2006
History: PJN / 02-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.     

Copyright (c) 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#ifndef __DTIMEPLUSCLDATE_H__
#define __DTIMEPLUSCLDATE_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////// Includes //////////////////////////////////////////

#include "DTime+Defines.h"
#include <cmath>


/////////////////////////////// Classes ///////////////////////////////////////

namespace DTimePlus
{

//A Date class with a granularity of 1 second
template <class TCALENDAR = CGregorianCalendar>
class CLDate
{
public:

//Constructors
  CLDate()
  {
    Set();
  }
  
  CLDate(long Year, WORD month, WORD Day, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {
    Set(Year, month, Day, Hour, Minute, Second, bDoAssert);
  }                                            

  CLDate(const SYSTEMTIME& SystemTime, bool bDoAssert = true)
  { 
    Set(SystemTime, bDoAssert);
  } 

  CLDate(const FILETIME& FileTime, bool bDoAssert = true)
  {
    Set(FileTime, bDoAssert);
  }

  CLDate(long Year, WORD month, WORD WeekOfMonth, DayOfWeek dow, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {
    Set(Year, month, WeekOfMonth, dow, Hour, Minute, Second, bDoAssert);
  }
               
  CLDate(long Days, DateEpoch e, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {
    Set(Days, e, Hour, Minute, Second, bDoAssert);
  } 

  CLDate(double Days, DateEpoch e, bool bDoAssert = true)
  {
    Set(Days, e, bDoAssert);
  } 

  CLDate(const CDate<TCALENDAR>& Date, const CLTimeOfDay<>& Tod)
  { 
    Set(Date, Tod);
  }

  CLDate(const CLDate& ld)
  {                                 
    m_nSeconds = ld.m_nSeconds;
  }

  CLDate(const __int64& nSeconds)
  {                                 
    m_nSeconds = nSeconds;
  }

  CLDate(const CTime& ctime, bool bDoAssert = true)
  {
    Set(ctime, bDoAssert);
  }
      
#ifndef DTIME_NO_OLE_SUPPORT
  CLDate(const COleDateTime& oleTime, bool bDoAssert = true)
  {
    Set(oleTime, bDoAssert);
  }                     
#endif

//Set operators
  CLDate& Set()
  {
    //Create an invalid value
    m_nSeconds = _I64_MIN;
    return *this;
  }

  CLDate& Set(long Year, WORD month, WORD Day, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {              
    CDate<TCALENDAR> Date(Year, month, Day, bDoAssert);
    CLTimeOfDay<> Tod(Hour, Minute, Second, bDoAssert);

    if (Date.IsValid() && Tod.IsValid())
    {
      m_nSeconds = static_cast<__int64>(Date.GetValue())*86400;
      if (m_nSeconds < 0)
        m_nSeconds -= Tod.GetTotalSeconds();
      else
        m_nSeconds += Tod.GetTotalSeconds();
    }
    else
      Set();

    return *this;
  }

  CLDate& Set(const SYSTEMTIME& st, bool bDoAssert = true)
  {
    Set(st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond, bDoAssert);

    return *this;
  }

  CLDate& Set(const FILETIME& FileTime, bool bDoAssert = true)
  {
    SYSTEMTIME SystemTime;
    if (::FileTimeToSystemTime(&FileTime, &SystemTime))
      Set(SystemTime, bDoAssert);
    else
    {
      if (bDoAssert)
        ASSERT(false);
      Set();
    }

    return *this;
  }

  CLDate& Set(long Year, WORD month, WORD WeekOfMonth, DayOfWeek dow, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {
    CDate<TCALENDAR> Date(Year, month, WeekOfMonth, dow, bDoAssert);
    CLTimeOfDay<> Tod(Hour, Minute, Second, bDoAssert);

    if (Date.IsValid() && Tod.IsValid())
      m_nSeconds = static_cast<_int64>(Date.GetValue())*86400 + Tod.GetTotalSeconds();
    else
      Set();

    return *this;
  }
                    
  CLDate& Set(long Days, DateEpoch e, WORD Hour, WORD Minute, WORD Second, bool bDoAssert = true)
  {
    CDate<TCALENDAR> Date(Days, e, bDoAssert);
    CLTimeOfDay<> Tod(Hour, Minute, Second, bDoAssert);

    if (Date.IsValid() && Tod.IsValid())
      m_nSeconds = static_cast<__int64>(Date.GetValue())*86400 + Tod.GetTotalSeconds();
    else
      Set();

    return *this;
  }

  CLDate& Set(double Days, DateEpoch e, bool bDoAssert = true)
  {              
    CDate<TCALENDAR> Date;
    if (Days < 0)
      Date.Set(static_cast<long>(Days-1), e, bDoAssert);
    else
      Date.Set(static_cast<long>(Days), e, bDoAssert);

    double dInt;
    double dFrac = modf(Days, &dInt);

    CLTimeOfDay<> Tod;
    if (dFrac < 0)
      Tod.Set(static_cast<DWORD>(86400 + dFrac*86400), bDoAssert);
    else
      Tod.Set(static_cast<DWORD>(dFrac*86400), bDoAssert);

    if (Date.IsValid() && Tod.IsValid())
    {
      m_nSeconds = static_cast<__int64>(Date.GetValue())*86400;
      if (m_nSeconds < 0)
        m_nSeconds -= Tod.GetTotalSeconds();
      else
        m_nSeconds += Tod.GetTotalSeconds();
    }
    else
      Set();

    return *this;
  }

  CLDate& Set(const __int64& nSeconds)
  {
    m_nSeconds = nSeconds;

    return *this;
  }

  CLDate& Set(const CTime& ctime, bool bDoAssert = true)
  {
    tm TM;
    if (ctime.GetGmtTm(&TM))
      Set(TM.tm_year + 1900, static_cast<WORD>(TM.tm_mon + 1), static_cast<WORD>(TM.tm_mday), static_cast<WORD>(TM.tm_hour), static_cast<WORD>(TM.tm_min), static_cast<WORD>(TM.tm_sec), bDoAssert);
    else
      Set();

    return *this;  
  }

  CLDate& Set(const CDate<TCALENDAR>& Date, const CLTimeOfDay<>& Tod)
  {
    if (!Date.IsValid() || !Tod.IsValid())
      Set();
    else
      m_nSeconds = static_cast<__int64>(Date.GetValue())*86400 + Tod.GetTotalSeconds();

    return *this;  
  }
      
#ifndef DTIME_NO_OLE_SUPPORT
  CLDate& Set(const COleDateTime& oleTime, bool bDoAssert = true)
  {
    COleDateTime::DateTimeStatus dts(oleTime.GetStatus());
    if (dts == COleDateTime::valid)
      Set(static_cast<long>(oleTime.GetYear()), static_cast<WORD>(oleTime.GetMonth()), static_cast<WORD>(oleTime.GetDay()), static_cast<WORD>(oleTime.GetHour()), static_cast<WORD>(oleTime.GetMinute()), static_cast<WORD>(oleTime.GetSecond()), bDoAssert);
    else
      Set();
    return *this;
  }                     
#endif

//Static constructors
  static CLDate CurrentLocalTime(bool bDoAssert = true)
  {               
    //PC Clock always return time in Gregorian Calendar
    SYSTEMTIME lt;
    ::GetLocalTime(&lt);
    CLDate<CGregorianCalendar> date(lt, bDoAssert);

    if (date.IsValid())
      return CLDate(date.GetValue()); 
    else
      return CLDate();
  }

  static CLDate CurrentSystemTime(bool bDoAssert = true)
  {
    //PC Clock always return time in Gregorian Calendar
    SYSTEMTIME st;
    ::GetSystemTime(&st);
    CLDate<CGregorianCalendar> date(st, bDoAssert);

    if (date.IsValid())
      return CLDate(date.GetValue());
    else
      return CLDate();
  }

//Operations
  CDate<TCALENDAR> GetCDate() const
  {
    ASSERT(IsValid());

    CDate<TCALENDAR> rVal;
    rVal.Set(static_cast<long>(m_nSeconds/86400), EPOCH_GREG);

    return rVal;
  }

  CLTimeOfDay<> GetCLTimeOfDay() const
  {
    ASSERT(IsValid());

    //What will be the return value from this function
    CLTimeOfDay<> rVal;

    __int64 nSeconds;
    if (m_nSeconds > 0)
      nSeconds = m_nSeconds - static_cast<__int64>(GetCDate().GetValue())*86400;
    else
    {
      __int64 nTempmSeconds = -(m_nSeconds);
      nSeconds = nTempmSeconds - static_cast<__int64>(-GetCDate().GetValue())*86400;
    }
    rVal.Set(static_cast<DWORD>(nSeconds));

    return rVal;
  }

  __int64 GetValue() const
  {
    ASSERT(IsValid());

    return m_nSeconds;
  }

  DateLS GetDate() const
  {
    ASSERT(IsValid());

    DateS ds(GetCDate().GetDate());

    //What will be the return value from this function
    DateLS rVal;

    rVal.lYear = ds.lYear; 
    rVal.wMonth = ds.wMonth;
    rVal.wDay = ds.wDay;
    rVal.weekday = ds.weekday;
    rVal.wYearday = ds.wYearday;
    CLTimeOfDay<> tod(GetCLTimeOfDay());
    rVal.wHour = tod.GetHours();
    rVal.wMinute = tod.GetMinutes();
    rVal.wSecond = tod.GetSeconds();
  
    return rVal;
  };

  bool GetAsSystemTime(SYSTEMTIME& SystemTime) const
  {
    if (!IsValid())
      return false;
  
    if (!GetCDate().GetAsSystemTime(SystemTime))
      return false;

    CLTimeOfDay<> tod(GetCLTimeOfDay());
    SystemTime.wHour = tod.GetHours();
    SystemTime.wMinute = tod.GetMinutes();
    SystemTime.wSecond = tod.GetSeconds();
    SystemTime.wMilliseconds = 0;

    return true;
  }

  bool GetAsFileTime(FILETIME& FileTime) const
  {
    SYSTEMTIME SystemTime;
    bool bSuccess = GetAsSystemTime(SystemTime);
    bSuccess = bSuccess && ::SystemTimeToFileTime(&SystemTime, &FileTime);
    return bSuccess;
  }

  bool GetAsTM(tm& time)
  {
    if (!IsValid())
     return false;

    //What will be the return value from this function
    bool rVal = true;

    DateLS ds(GetDate());

    long MinYear = 1900 + INT_MIN;
    long MaxYear = INT_MAX;
 
    //handle range errors
    if ((ds.lYear < MinYear) || (ds.lYear > MaxYear))
      rVal = false;
    else
    {
      time.tm_hour  = ds.wHour;
      time.tm_min   = ds.wMinute;
      time.tm_sec   = ds.wSecond;
      time.tm_year  = static_cast<int>(ds.lYear - 1900);
      time.tm_mon   = ds.wMonth - 1; //tm struct uses 0 based indices
      time.tm_wday  = ds.weekday - 1; //tm struct uses 0 based indices
      time.tm_mday  = ds.wDay;
      time.tm_yday  = ds.wYearday - 1; //Returns days since Jan 1
      time.tm_isdst = 0; //As you can see, GetTM always return 0 for this value. Client code should not rely on this value!
    }

    return rVal;
  }

#ifndef DTIME_NO_OLE_SUPPORT
  bool GetAsCOleDateTime(COleDateTime& OleDateTime) const
  {
    SYSTEMTIME st;
    bool rVal = GetAsSystemTime(st);
    if (rVal)
      OleDateTime = COleDateTime(st);
    else
      OleDateTime.SetStatus(COleDateTime::invalid);

    return rVal;
  }
#endif

  CLDate<CJulianCalendar> AsJulianCLDate() const
  {
    CLDate<CJulianCalendar> rVal;
    rVal.Set(m_nSeconds);

    return rVal;
  }

  CLDate<CGregorianCalendar> AsGregorianCLDate() const
  {
    CLDate<CGregorianCalendar> rVal;
    rVal.Set(m_nSeconds);

    return rVal;
  }

  void AddYear(int Years = 1)
  {
    ASSERT(IsValid());

    CDate<TCALENDAR> Date(GetCDate());
    CLTimeOfDay<> Tod(GetCLTimeOfDay());
    Date.AddYear(Years);
    Set(Date, Tod);
  }

  void AddMonth(int Months = 1)
  {
    ASSERT(IsValid());

    CDate<TCALENDAR> Date(GetCDate());
    CLTimeOfDay<> Tod(GetCLTimeOfDay());
    Date.AddMonth(Months);
    Set(Date, Tod);
  }

  void AddWeek(int Weeks = 1)
  {
    ASSERT(IsValid());

    CDate<TCALENDAR> Date(GetCDate());
    CLTimeOfDay<> Tod(GetCLTimeOfDay());
    Date.AddWeek(Weeks);
    Set(Date, Tod);
  }

  bool IsValid() const
  {
    return (m_nSeconds != _I64_MIN);
  }
                                      
//Overloaded Arithmetic Operators
  CLDate& operator=(const CLDate& d) 
  {
    m_nSeconds = d.m_nSeconds;
    return *this;
  }              

  CLDate operator+(const CLTimeSpan& TimeSpan) const
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      CLDate rVal;
      rVal.m_nSeconds = m_nSeconds + TimeSpan.GetValue();

      return rVal;
    }
    else 
      return CLDate();
  }

  CLTimeSpan operator-(CLDate& ld)
  {
    if (IsValid() && ld.IsValid())
      return CLTimeSpan(m_nSeconds - ld.m_nSeconds);
    else
      return CLTimeSpan();
  }

  CLDate operator-(const CLTimeSpan& TimeSpan)
  {
    if (IsValid() && TimeSpan.IsValid())
    {
      __int64 s = m_nSeconds - TimeSpan.GetValue();
      CLDate rVal;
      rVal.m_nSeconds = s;

      return rVal;
    }
    else
      return CLDate();
  }

  CLDate& operator+=(const CLTimeSpan& TimeSpan)
  {
    *this = *this + TimeSpan;
    return *this;
  }

  CLDate& operator-=(const CLTimeSpan& TimeSpan)
  {
    *this = *this - TimeSpan;
    return *this;
  }

  CLDate& operator++()
  {
    *this = *this + CLTimeSpan(1, 0, 0, 0);
    return *this;
  }

  CLDate& operator--()
  {
    *this = *this - CLTimeSpan(1, 0, 0, 0);
    return *this;
  }

//Overloaded Equality operators
  bool operator==(const CLDate& ld)
  {
    return (m_nSeconds == ld.m_nSeconds);
  }

  bool operator>(const CLDate& ld)
  {
    bool bValid = IsValid();
    bool bOtherValid = ld.IsValid();
    if (bValid && bOtherValid)
      return (m_nSeconds > ld.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator>=(const CLDate& ld)
  {
    return operator==(ld) || operator>(ld);
  }

  bool operator<(const CLDate& ld)
  {
    bool bValid = IsValid();
    bool bOtherValid = ld.IsValid();
    if (bValid && bOtherValid)
      return (m_nSeconds < ld.m_nSeconds);
    else
      return false; //Only makes sense to compare 2 valid values
  }

  bool operator<=(const CLDate& ld)
  {
    return operator==(ld) || operator<(ld);
  }

  bool operator!=(const CLDate& ld)
  {
    return !operator==(ld);
  }

//Display
  /*  //The Following Format parameters are supported
  %a  Abbreviated weekday name
  %A  Full weekday name
  %b  Abbreviated month name
  %B  Full month name
  %d  Day of month as decimal number (01 - 31)
  %j  Day of year as decimal number (001 - 366)
  %m  Month as decimal number (01 - 12)
  %P  AM / PM indicator
  %U  Week of year as decimal number
  %w  Weekday as decimal number (1 - 7; Sunday is 1)
  %x  Windows short date representation, namely Short Date representation + " " + Time Representation using GetDateFormat() and GetTimeFormat() Win32 API
  %y  Year without century, as decimal number (00 - 99)
  %Y  Year with century, as decimal number
  %H  Hours in the day
  %h  12 Hour format Hours in (00 - 12)
  %M  Minutes in the hour
  %S  Seconds in the minute
  %%  Percent sign        
  %#x Long date representation, namely Long Date representation + " " + Time Representation using GetDateFormat() and GetTimeFormat() Win32 API
  %#d, %#j, %#m, %#U, %#w, %#y, %#H, %#h, %#M, %#S  Remove leading zeros (if any).
  */
  CString Format(const CString& sFormat = GetWindowsFormatString(LOCALE_USER_DEFAULT), LCID Locale = LOCALE_USER_DEFAULT) const
  { 
    //What will be the return value from this function
    CString rVal;

    //Cache some values for speed optimization
    CDate<TCALENDAR> date(GetCDate());
    DateS ds = date.GetDate();
    CLTimeOfDay<> tod(GetCLTimeOfDay());

    if (IsValid())
    {
      CString sBuffer;
      int sFmtLength = sFormat.GetLength();

      for (int i=0; i<sFmtLength; i++)
      {
        TCHAR c = sFormat.GetAt(i);
        if (c == '%')
        {
          ++i;
          if (i < sFmtLength)
          {
            c = sFormat.GetAt(i);
            switch (c)
            {
              case _T('H'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(tod.GetHours()));
                rVal += sBuffer;
                break;
              }
              case _T('h'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(tod.GetAMPMHours()));
                rVal += sBuffer;
                break;
              }
              case _T('M'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(tod.GetMinutes()));
                rVal += sBuffer;
                break;
              }
              case _T('P'):
              {
                rVal += tod.GetAMPMString(Locale);
                break;
              }
              case _T('S'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(tod.GetSeconds()));
                rVal += sBuffer;
                break;
              }
              case _T('a'):
              {
                rVal += TCALENDAR::GetAbrStringDayOfWeek(ds.weekday, Locale);
                break;
              }
              case _T('A'):
              {
                rVal += TCALENDAR::GetFullStringDayOfWeek(ds.weekday, Locale);
                break;
              }
              case _T('b'):
              {
                rVal += TCALENDAR::GetAbrStringMonth(ds.wMonth, Locale);
                break;
              }
              case _T('B'):
              {
                rVal += TCALENDAR::GetFullStringMonth(ds.wMonth, Locale);
                break;
              }
              case _T('d'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(ds.wDay));
                rVal += sBuffer;
                break;
              }
              case _T('j'):
              {
                sBuffer.Format(_T("%.03d"), static_cast<int>(ds.wYearday));
                rVal += sBuffer;
                break;
              }
              case _T('m'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(ds.wMonth));
                rVal += sBuffer;
                break;
              }
              case _T('U'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(date.GetWeekOfYear(Locale)));
                rVal += sBuffer;
                break;
              }
              case _T('w'):
              {
                sBuffer.Format(_T("%d"), static_cast<int>(ds.weekday));
                rVal += sBuffer;
                break;
              }
              case _T('x'):
              {
                rVal += date.Format(_T("%x")) + _T(" ") + tod.Format(_T("%x"));
                break;
              }
              case _T('y'):
              {
                sBuffer.Format(_T("%.02d"), static_cast<int>(CDate<TCALENDAR>::Get2DigitYear(ds.lYear)));
                rVal += sBuffer;
                break;
              }
              case _T('Y'):
              {
                sBuffer.Format(_T("%ld"), ds.lYear);
                rVal += sBuffer;
                break;
              }
              case _T('#'):
              {
                if (i < sFmtLength)
                {
                  ++i;
                  c = sFormat.GetAt(i);
                  switch (c)
                  {
                    case _T('d'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wDay));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('j'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wYearday));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('m'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(ds.wMonth));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('U'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(date.GetWeekOfYear(Locale)));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('x'):
                    {
                      rVal += date.Format(_T("%#x")) + _T(" ") + tod.Format(_T("%x"));
                      break;
                    }
                    case _T('y'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(CDate<TCALENDAR>::Get2DigitYear(ds.lYear)));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('H'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(tod.GetHours()));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('h'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(tod.GetAMPMHours()));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('M'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(tod.GetMinutes()));
                      rVal += sBuffer;
                      break;
                    }
                    case _T('S'):
                    {
                      sBuffer.Format(_T("%d"), static_cast<int>(tod.GetSeconds()));
                      rVal += sBuffer;
                      break;
                    }
                    default:
                    {
                      rVal += c;
                      break;
                    }
                  }
                }
                break;
              }
              default:
              {
                rVal += c;
                break;
              }
            }
          }
        }
        else
        {
          rVal += c;
        }
      }
    }

    return rVal;
  }

//Serialization
  void Serialize(CArchive& ar)
  {
    if (ar.IsStoring())
    {
      WORD wVersion = 0x101;
      ar << wVersion;

      ar.Write(&m_nSeconds, sizeof(m_nSeconds));
    }
    else
    {
      WORD wVersion;
      ar >> wVersion;
      ASSERT(wVersion == 0x101 || wVersion == 0x100);

      ar.Read(&m_nSeconds, sizeof(m_nSeconds));
      if (wVersion == 0x100)
      {
        bool bOldInGregCalendar;
        ar.Read(&bOldInGregCalendar, sizeof(bOldInGregCalendar));
      }
    }
  }

  friend CArchive& operator<<(CArchive& ar, CLDate& ld)
  {
    ASSERT(ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }

  friend CArchive& operator>>(CArchive& ar, CLDate& ld)
  {
    ASSERT(!ar.IsStoring());
    ld.Serialize(ar);
    return ar;
  }

protected:
//Methods
  static CString GetWindowsFormatString(LCID Locale = LOCALE_USER_DEFAULT)
  {
    return CDate<TCALENDAR>::GetWindowsFormatString(Locale) + _T(" ") + CLTimeOfDay<>::GetWindowsFormatString(Locale);
  }

//Member variables
  __int64 m_nSeconds; //Actual instance data (count of seconds from same epoch as CDate uses
};

} //namespace DTimePlus

#endif //__DTIMEPLUSCLDATE_H__
