/*
Module : DTime+GregorianCalendar.H
Purpose: Defines the interface for the DTime+ CGregorianCalendar class
Created: PJN / 01-05-2006
History: PJN / 20-05-2006 1. CGregorianCalendar now uses functors for leap year calculations
         PJN / 29-05-2005 1. Further performance optimizations for the core CGregorianCalendar
                          algorithms.
         PJN / 02-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.

Copyright (c) 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#ifndef __DTIMEPLUSGREGORIANCALENDAR_H__
#define __DTIMEPLUSGREGORIANCALENDAR_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////// Includes //////////////////////////////////////////

#include "DTime+Defines.h"
#include "DTime+Helpers.h"
#include "DTime+JulianGregorianCalendar.h"


/////////////////////////// Classes ///////////////////////////////////////////

namespace DTimePlus
{

class CGregorianLeapFunctor
{
public:
  bool operator()(long Year) const
  {
    if ((Year % 100) == 0)                
      return ((Year % 400) == 0) ? true : false;
    else
      return ((Year % 4) == 0) ? true : false;
  }
};

//Class which implements the Gregorian calendar
class CGregorianCalendar : public CJulianGregorianCalendar<CGregorianLeapFunctor>
{
public:
  static DateS GetDate(long Days)
  {
    //What will be the return value from this function
    DateS ds;

    long gdn = Days + 577735;
    bool exception = false;
    ds.lYear = 400*lquotient(gdn, 146097);
    gdn -= 146097*lquotient(gdn, 146097); //400 year periods
    if (gdn > 0)
    {
      ds.lYear += 100*lquotient(gdn, 36524); //100 year periods
      gdn -= 36524*lquotient(gdn, 36524);
      exception = (gdn == 0);
      if (gdn > 0)
      {
        ds.lYear += 4*lquotient(gdn, 1461); //4 year periods
        gdn -= 1461*lquotient(gdn, 1461);
        if (gdn > 0)
        {
          int i=0;
          while (gdn > 365 && ++i < 4)
          {
            ds.lYear++;
            gdn -= 365;
          }
        }
      }
    }
    if (exception)
      gdn = 366; //occurs once every hundred years with Gregorian calendar
    else
    {
      ds.lYear++;
      gdn++;
    }

    static int MonthLength1[14] =
    { 
      0,  31, 28, 31, 30, 31, 30,
      31, 31, 30, 31, 30, 31, 0
    };

    static int MonthLength2[14] =
    { 
      0,  31, 29, 31, 30, 31, 30,
      31, 31, 30, 31, 30, 31, 0
    };
    int nMonth = January;
    CGregorianLeapFunctor pLeap;
    if (pLeap(ds.lYear))
    {
      while (nMonth <= December && gdn > MonthLength2[nMonth])
        gdn -= MonthLength2[nMonth++];
    }
    else
    {
      while (nMonth <= December && gdn > MonthLength1[nMonth])
        gdn -= MonthLength1[nMonth++];
    }
    if (nMonth == 13)
    {
      nMonth = January;
      ds.lYear++;
    }
    ds.wMonth = static_cast<WORD>(nMonth);
    ds.wDay = static_cast<WORD>(gdn);
    ds.weekday = static_cast<DayOfWeek>(lmod(Days+12, 7) + 1);
    ds.wYearday = DaysSinceEndPreviousYear(ds.lYear, ds.wMonth, ds.wDay);

    return ds;
  }

  inline static bool IsLeap(long Year)
  {
    return CGregorianLeapFunctor()(Year);
  }

  inline static void EasterSunday(long Year, WORD& month, WORD& Day)
  {
    int a = static_cast<int>(Year % 19);
    int b = static_cast<int>(Year / 100);
    int c = static_cast<int>(Year % 100);
    int d = b / 4;
    int e = b % 4;
    int f = (b+8) / 25;
    int g = (b - f + 1) / 3;
    int h = (19*a + b - d - g + 15) % 30;
    int i = c / 4;
    int k = c % 4;
    int l = (32 + 2*e + 2*i - h -k) % 7;
    int m = (a + 11*h +22*l) / 451;
    int n = (h + l - 7*m + 114) / 31;
    int p = (h + l - 7*m + 114) % 31;

    //Set the output parameters
    month = static_cast<WORD>(n);
    Day =  static_cast<WORD>(p+1);
  }

protected:
#ifdef _DEBUG
  static bool DMYToLong(long Year, WORD month, WORD Day, long& lDays, bool bDoAsserts)
#else
  static bool DMYToLong(long Year, WORD month, WORD Day, long& lDays, bool /*bDoAsserts*/)
#endif
  {
    //Validate our parameters
    if (month < January || month > December || Day > DaysInMonth(Year, month))
    {
    #ifdef _DEBUG
      if (bDoAsserts)
        ASSERT(false);
    #endif
      return false;
    }
    int nLocalMonth(month);

    static int MonthLength[14] =
    { 
      0,  31, 28, 31, 30, 31, 30,
      31, 31, 30, 31, 30, 31, 0
    };

    lDays = (Year-1)*365 + lquotient(Year-1, 4) + lquotient(Year-1, 400) - lquotient(Year-1, 100);

    --nLocalMonth;
    CGregorianLeapFunctor pLeap;
    while (nLocalMonth)
    {
      if (nLocalMonth == February && pLeap(Year))
        lDays += MonthLength[nLocalMonth] + 1;
      else
        lDays += MonthLength[nLocalMonth];
      nLocalMonth--;
    }

    lDays += Day - 577736;

    return TRUE;
  }
};

} //namespace DTimePlus

#endif //__DTIMEPLUSGREGORIANCALENDAR_H__
