/*
Module : DTime+HebrewCalendar.H
Purpose: Defines the interface for the DTime+ CHebrewCalendar class
Created: PJN / 01-05-2006
History: PJN / 25-05-2006 1. Updated the string and enums for the 7th month to be "Tishrei". Thanks to 
                          Itamar Syn-Hershko for reporting this issue.
                          2. Updated the string and enums for the first month to be "Nissan". Thanks to
                          Itamar Syn-Hershko for reporting this issue.
                          3. Updated the string and enums for the 8th month to be "Cheshvan". Thanks to 
                          Itamar Syn-Hershko for reporting this issue.
                          4. Now includes spaces in the strings returned for AdarI and AdarII.
                          5. Fixed a typo in the string and enums for the 11th month "Shevat". Thanks to 
                          Itamar Syn-Hershko for reporting this issue. 
                          6. Made some Hebrew functions inline to aid optimization
         PJN / 27-05-2006 1. Reworked hebrew algorithms to support dates before the Hebrew epoch i.e.
                          a proleptic hebrew calendar
                          2. Renamed the LongMarheshvan function to LongCheshvan.
                          3. Renamed the ShortMarheshvan function to ShortCheshvan 
         PJN / 02-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.

Copyright (c) 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#ifndef __DTIMEPLUSHEBREWCALENDAR_H__
#define __DTIMEPLUSHEBREWCALENDAR_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////// Includes //////////////////////////////////////////

#include "DTime+Defines.h"
#include "DTime+Helpers.h"


/////////////////////////// Classes ///////////////////////////////////////////

namespace DTimePlus
{

//Enums for the Hebrew Month names
enum HebrewMonth
{
  Nissan   = 1,
  Iyyar    = 2,
  Sivan    = 3,
  Tammuz   = 4,
  Av       = 5,
  Elul     = 6,
  Tishrei  = 7,
  Cheshvan = 8,
  Kislev   = 9,
  Tevet    = 10,
  Shevat   = 11,
  AdarI    = 12,
  AdarII   = 13
};

//Class which implements the Hebrew calendar
class CHebrewCalendar
{
public:
  inline static long HebrewNewYear(long Year)
  {
    static long HEBREW_EPOCH = -1951163; //Corresponds to 7 October 3761 B.C.E. O.S

    return HEBREW_EPOCH + HebrewCalendarElapsedDays(Year) + HebrewNewYearDelay(Year);
  }

  inline static WORD HebrewNewYearDelay(long Year)
  {
    long ny0 = HebrewCalendarElapsedDays(Year - 1);
    long ny1 = HebrewCalendarElapsedDays(Year);
    long ny2 = HebrewCalendarElapsedDays(Year + 1);

    if (ny2 - ny1 == 356)
      return 2;
    else if (ny1 - ny0 == 382)
      return 1;
    else
      return 0;
  }

  inline static long HebrewCalendarElapsedDays(long Year)
  {
    //This is the approach we would use if we wanted to avoid the use of floating point arithmetic below
    /*
    long nMonthsElapsed = lquotient(235*Year - 234, 19);
    __int64 nPartsElapsed = 12084 + 13753*(static_cast<__int64>(nMonthsElapsed));
    long nDay = static_cast<long>((29*nMonthsElapsed) + lquotient(nPartsElapsed, 25920));
    */

    static long HEBREW_EPOCH = -1951163; //Corresponds to 7 October 3761 B.C.E. O.S

    int nDay = static_cast<int>(Molad(Year, Tishrei) - HEBREW_EPOCH + 0.5);
    return (lmod(3*(nDay+1), 7) < 3) ? nDay + 1 : nDay;
  }

  inline static bool LongCheshvan(long Year)
  {
    WORD wDaysInYear = DaysInYear(Year);
    return (wDaysInYear == 355 || wDaysInYear == 385) ? true : false;
  }

  inline static bool ShortCheshvan(long Year)
  {
    return !LongCheshvan(Year);
  }

  inline static bool LongKislev(long Year)
  {
    return !ShortKislev(Year);
  }

  inline static bool ShortKislev(long Year)
  {
    WORD wDaysInYear = DaysInYear(Year);
    return (wDaysInYear == 353 || wDaysInYear == 383) ? true : false;
  }

  inline static double Molad(long Year, WORD month)
  {
    static long HEBREW_EPOCH = -1951163; //Corresponds to 7 October 3761 B.C.E. O.S

    long nMonthsElapsed = month - Tishrei + lquotient(235*Year - 234, 19);

    return HEBREW_EPOCH - 876.0/25920 + nMonthsElapsed*29.53059413580246913580246913580247;
  }

  inline static WORD DaysSinceEndPreviousYear(long Year, WORD month, WORD Day)
  { 
    //Validate our parameters           
    ASSERT(month >= Nissan && month <= LastMonthOfYear(IsLeap(Year)));
    ASSERT(Day >= 1 && Day <= DaysInMonth(Year, month));

    int nDays = 0;
    if (month < Tishrei)
    {
      WORD i;
      for (i=Tishrei; i<=LastMonthOfYear(IsLeap(Year)); i++)
        nDays += DaysInMonth(Year, i);
      for (i=Nissan; i<month; i++)
        nDays += DaysInMonth(Year, i);
    }
    else
    {
      for (WORD i=Tishrei; i<month; i++)
        nDays += DaysInMonth(Year, i);
    }
    nDays += Day;

    return static_cast<WORD>(nDays);
  }

  inline static WORD DaysSinceFirstOfYear(long Year, WORD month, WORD Day)
  {
    return static_cast<WORD>(DaysSinceEndPreviousYear(Year, month, Day) - 1);
  }

  inline static WORD DaysInMonth(long Year, WORD month)
  {
    if (month == Iyyar || month == Tammuz || month == Elul || month == Tevet || month == AdarII || (month == AdarI && !IsLeap(Year)) || 
        (month == Cheshvan && !LongCheshvan(Year)) || (month == Kislev && ShortKislev(Year)))
      return 29;
    else
      return 30;
  }

  inline static WORD DaysInYear(long Year)
  {
    return static_cast<WORD>(HebrewNewYear(Year+1) - HebrewNewYear(Year));
  }

  inline static WORD LastMonthOfYear(bool IsLeap)
  {
    return IsLeap ? static_cast<WORD>(13) : static_cast<WORD>(12);
  }

  static DateS GetDate(long Days)
  {
    static long HEBREW_EPOCH = -1951163; //Corresponds to 7 October 3761 B.C.E. O.S (JD 347997.5)

    //What will be the return value from this function
    DateS ds;
    memset(&ds, 0, sizeof(ds));

    long nApproxYear = static_cast<long>(1 + dquotient((Days - HEBREW_EPOCH), 35975351.0/98496));

    while (HebrewNewYear(nApproxYear) <= Days)
      ++nApproxYear;
    ds.lYear = nApproxYear - 1;

    long nTempDays;
    VERIFY(DMYToLong(ds.lYear, Nissan, 1, nTempDays, true));
    ds.wMonth = static_cast<WORD>((Days < nTempDays) ? Tishrei : Nissan);

    VERIFY(DMYToLong(ds.lYear, ds.wMonth, DaysInMonth(ds.lYear, ds.wMonth), nTempDays, true));
    while (nTempDays < Days)
    {
      ++ds.wMonth;
      VERIFY(DMYToLong(ds.lYear, ds.wMonth, DaysInMonth(ds.lYear, ds.wMonth), nTempDays, true));
    }

    VERIFY(DMYToLong(ds.lYear, ds.wMonth, 1, nTempDays, true));
    ds.wDay = static_cast<WORD>(Days - nTempDays + 1);

    long lFirstOfMonth;
    VERIFY(DMYToLong(ds.lYear, ds.wMonth, 1, lFirstOfMonth, true));
    ds.wDay = static_cast<WORD>(Days - lFirstOfMonth + 1);
    ds.weekday = static_cast<DayOfWeek>(lmod(Days+12, 7) + 1);
    ds.wYearday = DaysSinceEndPreviousYear(ds.lYear, ds.wMonth, ds.wDay);

    return ds;
  }

  inline static bool IsLeap(long Year)
  {
    return (lmod((7*Year + 1),19) < 7) ? true : false;
  }

  inline static CString GetFullStringDayOfWeek(DayOfWeek dow, LCID /*Locale*/)
  {
    //Validate our parameters           
    ASSERT(dow >= Sunday && dow <= Saturday);

    static TCHAR* pszHebrewDOW[] = 
    {
      _T("yom rishon"),
      _T("yom sheni"),
      _T("yom shelishi"), 
      _T("yom revi'i"), 
      _T("yom hamishi"),
      _T("yom shishi"), 
      _T("yom shabbat")
    };

    return pszHebrewDOW[dow - Sunday];
  }

  inline static CString GetAbrStringDayOfWeek(DayOfWeek dow, LCID Locale)
  {
    //There is not concept of an abreviated day of week name in the Hebrew calendar (That I 
    //know of), hence just delegate the work to GetFullStringDayOfWeek
    return GetFullStringDayOfWeek(dow, Locale);
  }

  inline static DayOfWeek GetBeginingDayOfWeek(LCID /*Locale*/)
  {
    //We do not use the Win32 NLS API's here, instead we just return the hardcoded value of Sunday
    return Sunday;
  }

  static CString GetFullStringMonth(WORD month, LCID /*Locale*/)
  {
    //Validate our parameters           
    ASSERT(month >= Nissan && month <= AdarII);

    static TCHAR* pszHebrewMonths[] = 
    {
      _T("Nissan"),
      _T("Iyyar"),
      _T("Sivan"),
      _T("Tammuz"),
      _T("Av"),
      _T("Elul"),
      _T("Tishrei"),
      _T("Cheshvan"),
      _T("Kislev"),
      _T("Tevet"),
      _T("Shevat"),
      _T("Adar I"),
      _T("Adar II")
    };

    return pszHebrewMonths[month - Nissan];
  }

  inline static CString GetAbrStringMonth(WORD month, LCID Locale)
  {
    //There is not concept of an abreviated Month name in the Hebrew calendar (That I 
    //know of), hence just delegate the work to GetFullStringMonth
    return GetFullStringMonth(month, Locale);
  }

protected:
#ifdef _DEBUG
  static bool DMYToLong(long Year, WORD month, WORD Day, long& lDays, bool bDoAssert)
#else
  static bool DMYToLong(long Year, WORD month, WORD Day, long& lDays, bool /*bDoAssert*/)
#endif
  {
    //Validate our parameters
    bool bLeap = IsLeap(Year);
    if (month < Nissan || month > LastMonthOfYear(bLeap) || Day > DaysInMonth(Year, month))
    {
    #ifdef _DEBUG
      if (bDoAssert)
        ASSERT(false);
    #endif
      return false;
    }

    lDays = HebrewNewYear(Year) + Day - 1;
    
    if (month < Tishrei)
    {
      WORD i;
      for (i=Tishrei; i<=LastMonthOfYear(bLeap); i++)
        lDays += DaysInMonth(Year, i);

      for (i=Nissan; i<month; i++)
        lDays += DaysInMonth(Year, i);
    }
    else
    {
      for (WORD i=Tishrei; i<month; i++)
        lDays += DaysInMonth(Year, i);
    }

    return true;
  }
};

} //namespace DTimePlus

#endif //__DTIMEPLUSHEBREWCALENDAR_H__
