/*
Module : DTime+JulianGregorianCalendar.H
Purpose: Defines the interface for the DTime+ CJulianGregorianCalendar class which is the base
         class for the CJulianCalendar and CGregorianCalendar classes
Created: PJN / 01-05-2006
History: PJN / 25-05-2006. 1. Optimized the code in CJulianGregorianCalendar::DaysInMonth
                           2. made all the functions in CJulianGregorianCalendar inline
         PJN / 02-07-2006 1. Updated the code to clean compile on VC 2005     
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.

Copyright (c) 2006 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////////////// Macros / Defines //////////////////////////////////

#ifndef __DTIMEPLUSJULIANGREGORIANCALENDAR_H__
#define __DTIMEPLUSJULIANGREGORIANCALENDAR_H__

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/////////////////////////// Includes //////////////////////////////////////////

#include "DTime+Defines.h"
#include "DTime+Helpers.h"


/////////////////////////// Classes ///////////////////////////////////////////

namespace DTimePlus
{

//Base class for the Julian and Greogiran calendars
template <class LEAP_YEAR_FUNCTOR>
class CJulianGregorianCalendar
{
public:
  inline static WORD DaysSinceEndPreviousYear(long Year, WORD month, WORD Day)
  { 
    //Validate our parameters           
    ASSERT(month >= January && month <= December);
    ASSERT(Day >= 1 && Day <= DaysInMonth(Year, month));

    LEAP_YEAR_FUNCTOR pLeap;
    int k = pLeap(Year) ? 1 : 2;
    WORD rVal = static_cast<WORD>((static_cast<int>(275*month/9)) - (k*static_cast<int>((month+9)/12)) + Day - 30);
    return rVal; 
  }

  inline static WORD DaysSinceFirstOfYear(long Year, WORD month, WORD Day)
  {
    return static_cast<WORD>(DaysSinceEndPreviousYear(Year, month, Day) - 1);
  }

  inline static WORD DaysInMonth(long Year, WORD month)
  {
    static WORD MonthLength[13] =
    { 
      0,  31, 28, 31, 30, 31, 30,
      31, 31, 30, 31, 30, 31
    };
    LEAP_YEAR_FUNCTOR pLeap;
    ASSERT(month >= January && month <= December);
    return (month == February && pLeap(Year)) ? static_cast<WORD>(MonthLength[month] + 1) : MonthLength[month];
  }

  inline static WORD DaysInYear(long Year)
  {
    LEAP_YEAR_FUNCTOR pLeap;
    return pLeap(Year) ? static_cast<WORD>(366) : static_cast<WORD>(365);
  }

  inline static DayOfWeek GetBeginingDayOfWeek(LCID Locale)
  {
    //What will be the return value from this function, assume Monday
    DayOfWeek dow(Monday);

    CString sBuf = SafeGetLocaleInfo(Locale, LOCALE_IFIRSTDAYOFWEEK);
    if (sBuf == _T("0"))
      dow = Monday;
    else if (sBuf == _T("1"))
      dow = Tuesday;
    else if (sBuf == _T("2"))
      dow = Wednesday;
    else if (sBuf == _T("3"))
      dow = Thursday;
    else if (sBuf == _T("4"))
      dow = Friday;
    else if (sBuf == _T("5"))
      dow = Saturday;
    else if (sBuf == _T("6"))
      dow = Sunday;
    else
      ASSERT(false);

    return dow;
  }

  inline static CString GetFullStringDayOfWeek(DayOfWeek dow, LCID Locale)
  {
    //What will be the return value from this function
    CString rVal;

    switch (dow)
    {
      case Monday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME1); 
        break;
      }
      case Tuesday:   
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME2); 
        break;
      }
      case Wednesday: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME3); 
        break;
      }
      case Thursday:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME4); 
        break;
      }
      case Friday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME5); 
        break;
      }
      case Saturday:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME6); 
        break;
      }
      case Sunday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SDAYNAME7); 
        break;
      }
      default: 
      {
        ASSERT(false); 
        break;
      }
    }

    return rVal;
  }

  inline static CString GetAbrStringDayOfWeek(DayOfWeek dow, LCID Locale)
  {
    //What will be the return value from this function
    CString rVal;

    switch (dow)
    {
      case Monday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME1); 
        break;
      }
      case Tuesday:   
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME2); 
        break;
      }
      case Wednesday: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME3); 
        break;
      }
      case Thursday:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME4); 
        break;
      }
      case Friday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME5); 
        break;
      }
      case Saturday:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME6); 
        break;
      }
      case Sunday:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVDAYNAME7); 
        break;
      }
      default: 
      {
        ASSERT(false); 
        break;
      }
    }

    return rVal;
  }

  inline static CString GetFullStringMonth(WORD month, LCID Locale)
  {
    //What will be the return value from this function
    CString rVal;

    switch (month)
    {
      case January:   
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME1); 
        break;
      }
      case February:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME2); 
        break;
      }
      case March: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME3); 
        break;
      }
      case April: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME4); 
        break;
      }
      case May: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME5); 
        break;
      }
      case June:      
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME6); 
        break;
      }
      case July:      
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME7); 
        break;
      }
      case August:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME8); 
        break;
      }
      case September: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME9); 
        break;
      }
      case October:   
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME10); 
        break;
      }
      case November:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME11); 
        break;
      }
      case December:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SMONTHNAME12); 
        break;
      }
      default: 
      {
        ASSERT(false); 
        break;
      }
    }

    return rVal;
  }

  inline static CString GetAbrStringMonth(WORD month, LCID Locale)
  {
    //What will be the return value from this function
    CString rVal;

    switch (month)
    {
      case January: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME1); 
        break;
      }
      case February:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME2); 
        break;
      }
      case March:     
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME3); 
        break;
      }
      case April:     
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME4); 
        break;
      }
      case May:       
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME5); 
        break;
      }
      case June:      
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME6); 
        break;
      }
      case July:      
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME7); 
        break;
      }
      case August:    
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME8); 
        break;
      }
      case September: 
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME9); 
        break;
      }
      case October:   
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME10); 
        break;
      }
      case November:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME11); 
        break;
      }
      case December:  
      {
        rVal = SafeGetLocaleInfo(Locale, LOCALE_SABBREVMONTHNAME12); 
        break;
      }
      default: 
      {
        ASSERT(false); 
        break;
      }
    }

    return rVal;
  }
};

} //namespace DTimePlus

#endif //__DTIMEPLUSJULIANGREGORIANCALENDAR_H__
